@@ -5,6 +5,13 @@ from django.contrib.auth.models import Group as DjangoGroup
 from django.db import models
 from django.db.models.signals import post_save, post_delete
 
+try:
+    # Pre Django 1.7 support
+    from django.db.models import get_model as django_get_model
+except ImportError:
+    from django.apps import apps
+    django_get_model = apps.get_model
+
 from django.conf import settings as django_settings
 from django.contrib.auth.models import User as DefaultUser
 DjangoUser = getattr(django_settings, 'AUTH_USER_MODEL', DefaultUser)
@@ -13,9 +20,13 @@ from jsonfield import JSONField
 from mptt.models import MPTTModel, TreeForeignKey
 from slugify import slugify
 
-import exceptions
+from groups_manager import exceptions
+from groups_manager.perms import assign_object_to_member, assign_object_to_group
 
-from .perms import assign_object_to_member
+
+def get_auth_models_sync_func_default(instance):
+    from groups_manager.settings import GROUPS_MANAGER
+    return GROUPS_MANAGER['AUTH_MODELS_SYNC']
 
 
 class Member(models.Model):
@@ -45,6 +56,9 @@ class Member(models.Model):
     def __unicode__(self):
         return self.full_name
 
+    def __str__(self):
+        return self.full_name
+
     def save(self, *args, **kwargs):
         if not self.username:
             self.username = slugify(self.full_name, to_lower=True, separator="_")
@@ -88,13 +102,17 @@ class Member(models.Model):
 
 
 def member_save(sender, instance, created, *args, **kwargs):
-    '''
+    """
     Add User to Django Users
-    '''
-    from settings import GROUPS_MANAGER
-    if GROUPS_MANAGER['AUTH_MODELS_SYNC'] and instance.django_auth_sync:
-        prefix = GROUPS_MANAGER['USER_USERNAME_PREFIX']
-        suffix = GROUPS_MANAGER['USER_USERNAME_SUFFIX']
+    """
+    from groups_manager.settings import GROUPS_MANAGER
+    get_auth_models_sync_func = kwargs.get('get_auth_models_sync_func',
+                                           get_auth_models_sync_func_default)
+    if get_auth_models_sync_func(instance) and instance.django_auth_sync:
+        original_prefix = kwargs.get('prefix', GROUPS_MANAGER['USER_USERNAME_PREFIX'])
+        original_suffix = kwargs.get('suffix', GROUPS_MANAGER['USER_USERNAME_SUFFIX'])
+        prefix = original_prefix
+        suffix = original_suffix
         if suffix == '_$$random':
             suffix = '_%s' % str(uuid4())[:8]
         username = '%s%s%s' % (prefix, instance.username, suffix)
@@ -111,10 +129,9 @@ def member_save(sender, instance, created, *args, **kwargs):
             instance.django_user = django_user
             instance.save()
         else:
-            if (instance.django_user.username != username and
-                suffix != '_$$random') \
+            if (instance.django_user.username != username and suffix != '_$$random') \
                 or (instance.django_user.username[:-len(suffix)] != username[:-len(suffix)] and
-                    GROUPS_MANAGER['USER_USERNAME_SUFFIX'] == '_$$random'):
+                    original_suffix == '_$$random'):
                 instance.django_user.username = username
             instance.django_user.first_name = instance.first_name
             instance.django_user.last_name = instance.last_name
@@ -123,11 +140,12 @@ def member_save(sender, instance, created, *args, **kwargs):
 
 
 def member_delete(sender, instance, *args, **kwargs):
-    '''
+    """
     Remove the related Django Group
-    '''
-    from settings import GROUPS_MANAGER
-    if GROUPS_MANAGER['AUTH_MODELS_SYNC'] and instance.django_auth_sync:
+    """
+    get_auth_models_sync_func = kwargs.get('get_auth_models_sync_func',
+                                           get_auth_models_sync_func_default)
+    if get_auth_models_sync_func(instance) and instance.django_auth_sync:
         if instance.django_user:
             django_user = instance.django_user
             django_user.delete()
@@ -155,6 +173,9 @@ class GroupType(models.Model):
     def __unicode__(self):
         return self.label
 
+    def __str__(self):
+        return self.label
+
     def save(self, *args, **kwargs):
         if not self.codename:
             self.codename = slugify(self.label, to_lower=True)
@@ -179,6 +200,9 @@ class GroupEntity(models.Model):
     def __unicode__(self):
         return self.label
 
+    def __str__(self):
+        return self.label
+
     def save(self, *args, **kwargs):
         if not self.codename:
             self.codename = slugify(self.label, to_lower=True)
@@ -211,7 +235,7 @@ class Group(MPTTModel):
      mptt.models.TreeManager.
 
     """
-    name = models.CharField(max_length=255, unique=True)
+    name = models.CharField(max_length=255)
     codename = models.SlugField(blank=True, max_length=255)
     description = models.TextField(default='', blank=True)
     comment = models.TextField(default='', blank=True)
@@ -223,8 +247,7 @@ class Group(MPTTModel):
 
     group_type = models.ForeignKey(GroupType, null=True, blank=True, on_delete=models.SET_NULL,
                                    related_name='groups')
-    group_entities = models.ManyToManyField(GroupEntity, null=True, blank=True,
-                                            related_name='groups')
+    group_entities = models.ManyToManyField(GroupEntity, blank=True, related_name='groups')
 
     django_group = models.ForeignKey(DjangoGroup, null=True, blank=True, on_delete=models.SET_NULL)
     django_auth_sync = models.BooleanField(default=True, blank=True)
@@ -236,9 +259,16 @@ class Group(MPTTModel):
         level_attr = 'level'
         order_insertion_by = ['name', ]
 
+    class GroupsManagerMeta:
+        member_model = 'groups_manager.Member'
+        group_member_model = 'groups_manager.GroupMember'
+
     def __unicode__(self):
         return '%s' % self.name
 
+    def __str__(self):
+        return '%s' % self.name
+
     def save(self, *args, **kwargs):
         self.full_name = self._get_full_name()[:255]
         if not self.codename:
@@ -250,13 +280,41 @@ class Group(MPTTModel):
             return '%s - %s' % (self.parent._get_full_name(), self.name)
         return self.name
 
+    @property
+    def member_model(self):
+        member_model_path = getattr(self.GroupsManagerMeta,
+                                    'member_model', 'groups_manager.Member')
+        return django_get_model(*member_model_path.split('.'))
+
+    @property
+    def group_member_model(self):
+        group_member_model_path = getattr(self.GroupsManagerMeta,
+                                          'group_member_model', 'groups_manager.GroupMember')
+        return django_get_model(*group_member_model_path.split('.'))
+
     def get_members(self, subgroups=False):
         """Return group members.
+        The result is a list of GroupsManagerMeta's attribute ``member_model`` instances.
 
         :Parameters:
           - `subgroups`: return also descendants members (default: `False`)
         """
-        members = list(self.group_members.all())
+        member_model = self.member_model
+        group_member_model = self.group_member_model
+        if group_member_model == GroupMember:
+            if member_model == Member:
+                members = list(self.group_members.all())
+            else:
+                # proxy model
+                if member_model._meta.proxy:
+                    members = list(member_model.objects.filter(
+                        id__in=self.group_members.values_list('id', flat=True)))
+                # subclassed
+                else:
+                    members = list(member_model.objects.filter(
+                        member_ptr__in=self.group_members.all()))
+        else:
+            members = [gm.member for gm in group_member_model.objects.filter(group=self)]
         if subgroups:
             for subgroup in self.subgroups.all():
                 members += subgroup.members
@@ -305,11 +363,12 @@ class Group(MPTTModel):
         """
         if not self.id:
             raise exceptions.GroupNotSavedError(
-                    "You must save the group before to create a relation with members")
+                "You must save the group before to create a relation with members")
         if not member.id:
             raise exceptions.MemberNotSavedError(
-                    "You must save the member before to create a relation with groups")
-        group_member = GroupMember(member=member, group=self)
+                "You must save the member before to create a relation with groups")
+        group_member_model = self.group_member_model
+        group_member = group_member_model(member=member, group=self)
         group_member.save()
         if roles:
             for role in roles:
@@ -327,6 +386,19 @@ class Group(MPTTModel):
                         raise exceptions.GetRoleError(e)
         return group_member
 
+    def remove_member(self, member):
+        """Remove a member from the group.
+
+        :Parameters:
+          - `member`: member (required)
+        """
+        group_member_model = self.group_member_model
+        try:
+            group_member = group_member_model.objects.get(member=member, group=self)
+        except Exception as e:
+            raise exceptions.GetGroupMemberError(e)
+        group_member.delete()
+
     def assign_object(self, obj, **kwargs):
         """Assign an object to the group.
 
@@ -338,22 +410,23 @@ class Group(MPTTModel):
 
         .. note::
          This method needs django-guardian.
-
-        .. todo::
-         Create a method that allows to assign_objects directly to group.
         """
-        raise NotImplementedError("This method has not yet implemented")
+        return assign_object_to_group(self, obj, **kwargs)
 
 
 def group_save(sender, instance, created, *args, **kwargs):
-    '''
+    """
     Add Group to Django Groups
-    '''
-    from settings import GROUPS_MANAGER
-    if GROUPS_MANAGER['AUTH_MODELS_SYNC'] and instance.django_auth_sync:
+    """
+    from groups_manager.settings import GROUPS_MANAGER
+    get_auth_models_sync_func = kwargs.get('get_auth_models_sync_func',
+                                           get_auth_models_sync_func_default)
+    if get_auth_models_sync_func(instance) and instance.django_auth_sync:
         # create a name compatible with django group name limit of 80 chars
-        prefix = GROUPS_MANAGER['GROUP_NAME_PREFIX']
-        suffix = GROUPS_MANAGER['GROUP_NAME_SUFFIX']
+        original_prefix = kwargs.get('prefix', GROUPS_MANAGER['GROUP_NAME_PREFIX'])
+        original_suffix = kwargs.get('suffix', GROUPS_MANAGER['GROUP_NAME_SUFFIX'])
+        prefix = original_prefix
+        suffix = original_suffix
         if suffix == '_$$random':
             suffix = '_%s' % str(uuid4())[:8]
         parent_name = ''
@@ -365,20 +438,20 @@ def group_save(sender, instance, created, *args, **kwargs):
             django_group.save()
             instance.django_group = django_group
             instance.save()
-        elif (instance.django_group.name != name and
-              GROUPS_MANAGER['GROUP_NAME_SUFFIX'] != '_$$random') \
+        elif (instance.django_group.name != name and original_suffix != '_$$random') \
                 or (instance.django_group.name[:-len(suffix)] != name[:-len(suffix)] and
-                    GROUPS_MANAGER['GROUP_NAME_SUFFIX'] == '_$$random'):
+                    original_suffix == '_$$random'):
             instance.django_group.name = name
             instance.django_group.save()
 
 
 def group_delete(sender, instance, *args, **kwargs):
-    '''
+    """
     Remove the related Django Group
-    '''
-    from settings import GROUPS_MANAGER
-    if GROUPS_MANAGER['AUTH_MODELS_SYNC'] and instance.django_auth_sync:
+    """
+    get_auth_models_sync_func = kwargs.get('get_auth_models_sync_func',
+                                           get_auth_models_sync_func_default)
+    if get_auth_models_sync_func(instance) and instance.django_auth_sync:
         if instance.django_group:
             django_group = instance.django_group
             django_group.delete()
@@ -406,6 +479,9 @@ class GroupMemberRole(models.Model):
     def __unicode__(self):
         return self.label
 
+    def __str__(self):
+        return self.label
+
     def save(self, *args, **kwargs):
         if not self.codename:
             self.codename = slugify(self.label, to_lower=True)
@@ -424,7 +500,7 @@ class GroupMember(models.Model):
     """
     group = models.ForeignKey(Group, related_name='group_membership')
     member = models.ForeignKey(Member, related_name='group_membership')
-    roles = models.ManyToManyField(GroupMemberRole, null=True, blank=True)
+    roles = models.ManyToManyField(GroupMemberRole, blank=True)
 
     class Meta:
         ordering = ('group', 'member')
@@ -433,6 +509,9 @@ class GroupMember(models.Model):
     def __unicode__(self):
         return '%s - %s' % (self.group.name, self.member.full_name)
 
+    def __str__(self):
+        return '%s - %s' % (self.group.name, self.member.full_name)
+
     def assign_object(self, obj, **kwargs):
         """Assign an object to the member.
 
@@ -449,11 +528,12 @@ class GroupMember(models.Model):
 
 
 def group_member_save(sender, instance, created, *args, **kwargs):
-    '''
+    """
     Add Django User to Django Groups
-    '''
-    from settings import GROUPS_MANAGER
-    if GROUPS_MANAGER['AUTH_MODELS_SYNC']:
+    """
+    get_auth_models_sync_func = kwargs.get('get_auth_models_sync_func',
+                                           get_auth_models_sync_func_default)
+    if get_auth_models_sync_func(instance):
         django_user = instance.member.django_user
         django_group = instance.group.django_group
         if django_user and django_group:
@@ -462,16 +542,24 @@ def group_member_save(sender, instance, created, *args, **kwargs):
 
 
 def group_member_delete(sender, instance, *args, **kwargs):
-    '''
+    """
     Remove Django User from Django Groups
-    '''
-    from settings import GROUPS_MANAGER
-    if GROUPS_MANAGER['AUTH_MODELS_SYNC']:
-        django_user = instance.member.django_user
-        django_group = instance.group.django_group
-        if django_user and django_group:
-            if django_group in django_user.groups.all():
-                django_user.groups.remove(django_group)
+    """
+    get_auth_models_sync_func = kwargs.get('get_auth_models_sync_func',
+                                           get_auth_models_sync_func_default)
+    if get_auth_models_sync_func(instance):
+        member = instance.member
+        # Django 1.4 compatibility
+        try:
+            group = instance.group
+        except Group.DoesNotExist:
+            group = None
+        if member and group:
+            django_user = instance.member.django_user
+            django_group = instance.group.django_group
+            if django_user and django_group:
+                if django_group in django_user.groups.all():
+                    django_user.groups.remove(django_group)
 
 post_save.connect(group_member_save, sender=GroupMember)
 post_delete.connect(group_member_delete, sender=GroupMember)
